///////////////////////////////////////////////////////////////////////////////
// squareVulkanized.cpp
//
// This program is square.cpp rewritten using the Vulkan API instead of OpenGL. 
//
// Configuring the environment to run this project: Follow the instructions at
// https://vulkan-tutorial.com/Development_environment
//
// Sumanta Guha
///////////////////////////////////////////////////////////////////////////////

#define GLFW_INCLUDE_VULKAN
#include <GLFW/glfw3.h>

#include <glm/glm.hpp>
#include <glm/gtc/matrix_transform.hpp>

#include <iostream>
#include <fstream>
#include <vector>
#include <cstring>

using namespace std;
using namespace glm;

struct Vertex 
{
   vec4 coords;
   vec4 colors;
};

struct MatrixUniforms 
{
   mat4 modelviewMat;
   mat4 projMat;
};

/* Globals begin. */
static int windowWidth = 500;
static int windowHeight = 500;

static Vertex squareVertices[] =
{
   { vec4(20.0, 20.0, 0.0, 1.0), vec4(0.0, 0.0, 0.0, 1.0) },
   { vec4(80.0, 20.0, 0.0, 1.0), vec4(0.0, 0.0, 0.0, 1.0) },
   { vec4(20.0, 80.0, 0.0, 1.0), vec4(0.0, 0.0, 0.0, 1.0) },
   { vec4(80.0, 80.0, 0.0, 1.0), vec4(0.0, 0.0, 0.0, 1.0) }
};

static MatrixUniforms matrixUniforms;

static GLFWwindow* window;

static VkInstance instance;
static VkSurfaceKHR surface;

static VkPhysicalDevice physicalDevice;
static VkDevice logicalDevice;
static int queueFamily;
static VkQueue queue;

static VkSwapchainKHR swapchain;
static vector<VkImage> swapchainImages;
static VkSurfaceFormatKHR  swapchainImageFormat;
static VkExtent2D swapchainImageExtent;
static vector<VkImageView> swapchainImageViews;

static VkPipelineLayout pipelineLayout;
static VkPipeline graphicsPipeline;
static VkRenderPass renderpass;
static vector<VkFramebuffer> framebuffers;

static VkCommandPool commandPoolDraw;
static VkCommandPool commandPoolCopy;
static vector<VkCommandBuffer> commandBuffersDraw;
static VkCommandBuffer commandBufferCopy;

static VkDescriptorPool descriptorPool;
static VkDescriptorSetLayout descriptorSetLayout;
static VkDescriptorSet descriptorSet;

static VkFence imageAcquiredFence;
static VkSemaphore renderCompleteSemaphore;

static VkBuffer vertexBuffer, vertexStagingBuffer, uniformBuffer;
static VkDeviceMemory vertexBufferMemory, vertexStagingBufferMemory, uniformBufferMemory;
/* Globals end. */

// Create a Vulkan instance.
void createInstance()
{
   // App info, i.e, info about this particular program (we state a bare minimum).
   VkApplicationInfo appInfo = {};
   appInfo.sType = VK_STRUCTURE_TYPE_APPLICATION_INFO;
   appInfo.pApplicationName = "squareVulkanized.cpp";

   // Instance info starts here.
   VkInstanceCreateInfo createInfo = {};
   createInfo.sType = VK_STRUCTURE_TYPE_INSTANCE_CREATE_INFO;

   // Pointer to app info from instance. 
   createInfo.pApplicationInfo = &appInfo;

   // Get the extensions GLFW requires to create a Vulkan surface.
   unsigned int reqdExtensionsCount = 0;
   const char** reqdExtensions;
   reqdExtensions = glfwGetRequiredInstanceExtensions(&reqdExtensionsCount);

   // We simply enable the reqdExtensions in our Vulkan instance 
   // without checking if they are indeed supported.
   // Almost always these extensions will be supported;
   // however, if not, then instance creation will simply fail.
   // Or we could first check which extensions are supported using
   // vkEnumerateInstanceExtensionProperties() and exit gracefully 
   // if all reqdExtensions are not supported.
   createInfo.enabledExtensionCount = reqdExtensionsCount;
   createInfo.ppEnabledExtensionNames = reqdExtensions;

   // Create instance.
   vkCreateInstance(&createInfo, NULL, &instance);
}

// Create surface.
void createSurface()
{
   glfwCreateWindowSurface(instance, window, NULL, &surface);
}

// Check if a given physical device is a separate GPU
// (we don't really care but are just trying to make the device selection procedure non-trivial).
// DISABLE THIS CHECK IF YOU DON'T HAVE A SEPARATE GPU. SEE pickPhysicalDeviceAndQueueFamily() BELOW.
bool devIsADiscreteGPU(VkPhysicalDevice device)
{
   VkPhysicalDeviceProperties devProperties;
   vkGetPhysicalDeviceProperties(device, &devProperties);

   return (devProperties.deviceType == VK_PHYSICAL_DEVICE_TYPE_DISCRETE_GPU);
}

// Check if a given physical device supports tessellation and geometry shaders
// (again, we don't need such shaders for this program but are just trying to make
// the device selection process non-trivial).
// The features will actually be enabled when we create the logical device.
bool devSupportsTessAndGeomShaders(VkPhysicalDevice device)
{
   VkPhysicalDeviceFeatures devFeatures;
   vkGetPhysicalDeviceFeatures(device, &devFeatures);

   return devFeatures.geometryShader && devFeatures.tessellationShader;
}

// Check if a given physical device has an extension supporting swap chains.
// The extension will actually be enabled later when creating the logical device.
bool devSupportsSwapchains(VkPhysicalDevice device)
{
   // Retrieve supported extensions.
   uint32_t devExtensionsCount;
   vkEnumerateDeviceExtensionProperties(device, NULL, &devExtensionsCount, NULL);
   vector<VkExtensionProperties> devExtensions(devExtensionsCount);
   vkEnumerateDeviceExtensionProperties(device, NULL, &devExtensionsCount, devExtensions.data());

   // Look for swapchain extension in supported extensions.
   for (auto extension : devExtensions)
	  if (strcmp(VK_KHR_SWAPCHAIN_EXTENSION_NAME, extension.extensionName) == 0) return true;

   return false;
}

// Given a physical device run through the device queue families checking for both 
// support for graphics commands and for presentation to the surface: return the 
// index of a suitable family if found, or -1 if not (note: we could relax the 
// requirement by accepting separate queue families supporting presentation and
// graphics ops, but we keep it simple asking one dual-purpose queue family).
int devQueueFamilyIndexSupGrAndPr(VkPhysicalDevice device)
{
   // Retrieve queue families.
   uint32_t devQueueFamilyCount = 0;
   vkGetPhysicalDeviceQueueFamilyProperties(device, &devQueueFamilyCount, NULL);
   vector<VkQueueFamilyProperties> devQueueFamilies(devQueueFamilyCount);
   vkGetPhysicalDeviceQueueFamilyProperties(device, &devQueueFamilyCount, devQueueFamilies.data());

   int i = 0;
   VkBool32 presentationSupport;

   // Run through queue families checking for graphics and presentation support.
   for (auto family : devQueueFamilies)
   {
	  presentationSupport = false;
	  vkGetPhysicalDeviceSurfaceSupportKHR(device, i, surface, &presentationSupport);
	  if ((family.queueFlags & VK_QUEUE_GRAPHICS_BIT)  // Queue family supports graphics ...
		 && presentationSupport // ... and presentation. 
		 ) return i;
	  i++;
	  if (i > devQueueFamilyCount - 1) return -1;
   }
}

// Put in all 4 checks above to find a physical device which is a discrete GPU
// supporting tessellation and geometry shaders and which has at least one
// queue family supporting graphics and presentation.
void pickPhysicalDeviceAndQueueFamily()
{
   // Retrieve available physical devices.
   uint32_t physicalDevsCount = 0;
   vkEnumeratePhysicalDevices(instance, &physicalDevsCount, NULL);
   vector<VkPhysicalDevice> physicalDevs(physicalDevsCount);
   vkEnumeratePhysicalDevices(instance, &physicalDevsCount, physicalDevs.data());

   int i = 0;

   // Run through physical devices looking for a fit.
   for (auto device : physicalDevs)
   {
	  int index = devQueueFamilyIndexSupGrAndPr(device);
	  if (
		 devIsADiscreteGPU(device)  // Device is a discrete GPU (comment out if you don't have a separate GPU) ... 
		 && devSupportsTessAndGeomShaders(device) // ... supporting tessellation and geometry shaders ...
		 && devSupportsSwapchains(device) // ... and swapchains ...
		 && (index >= 0) // ... with a queue family supporting graphics and presentation. 
		 )
	  {
		 physicalDevice = device;
		 queueFamily = index;
		 return;
	  }
	  i++;
	  if (i > physicalDevsCount - 1)
	  {
		 cout << "Failed to pick device!" << endl;
		 exit(EXIT_FAILURE);
	  }
   }
}

// Create logical device and get queue handle.
void createLogicalDeviceAndPickQueue()
{
   // Logical device's queue info.
   VkDeviceQueueCreateInfo queueCreateInfo = {};
   queueCreateInfo.sType = VK_STRUCTURE_TYPE_DEVICE_QUEUE_CREATE_INFO;
   queueCreateInfo.queueFamilyIndex = queueFamily;
   queueCreateInfo.queueCount = 1;
   float queuePriority = 1.0;
   queueCreateInfo.pQueuePriorities = &queuePriority;

   // Logical device info.
   VkDeviceCreateInfo createInfo = {};
   createInfo.sType = VK_STRUCTURE_TYPE_DEVICE_CREATE_INFO;
   createInfo.queueCreateInfoCount = 1;
   createInfo.pQueueCreateInfos = &queueCreateInfo;

   // As promised earlier the swap chain extension is now enabled.
   createInfo.enabledExtensionCount = 1;
   const char* const extenstionsToEnable = { VK_KHR_SWAPCHAIN_EXTENSION_NAME };
   createInfo.ppEnabledExtensionNames = &extenstionsToEnable;

   // As promised earlier two device features are now enabled.
   VkPhysicalDeviceFeatures deviceFeatures = {};
   deviceFeatures.geometryShader = true;
   deviceFeatures.tessellationShader = true;
   createInfo.pEnabledFeatures = &deviceFeatures;

   // Create logical device.
   vkCreateDevice(physicalDevice, &createInfo, NULL, &logicalDevice);

   // Get the handle to the one queue we want from the chosen queue family.
   vkGetDeviceQueue(logicalDevice, queueFamily, 0, &queue);
}

// Create the swapchain and image views, one per swapchain image.
void createSwapchainAndImageViews()
{
   // Retrieve supported surface formats.
   uint32_t surfFormatsCount;
   vkGetPhysicalDeviceSurfaceFormatsKHR(physicalDevice, surface, &surfFormatsCount, NULL);
   vector<VkSurfaceFormatKHR> surfFormats(surfFormatsCount);
   vkGetPhysicalDeviceSurfaceFormatsKHR(physicalDevice, surface, &surfFormatsCount, surfFormats.data());

   // Choose a format for the swapchain images:
   // If only one format VK_FORMAT_UNDEFINED is listed for the surface then user is free to choose any VkFormat value.
   // In which case we'll choose the popular VK_FORMAT_B8G8R8A8_UNORM format where each channel,
   // in the order BGRA, is an 8-bit unsigned integer.
   // As for the color space, at this time the only choice is sRGB, which determines how the
   // presentation surface interprets the color channel values as actual light intensities.
   if (surfFormats.size() == 1 && (surfFormats[0].format == VK_FORMAT_UNDEFINED))
	  swapchainImageFormat = { VK_FORMAT_B8G8R8A8_UNORM, VK_COLOR_SPACE_SRGB_NONLINEAR_KHR };

   // If there is more than one format then look for the VK_FORMAT_B8G8R8A8_UNORM format and
   // simply revert to the first format in the list if it's not found.
   else
   {
	  swapchainImageFormat = surfFormats[0];
	  for (auto aFormat : surfFormats)
		 if (aFormat.format == VK_FORMAT_B8G8R8A8_UNORM && aFormat.colorSpace == VK_COLOR_SPACE_SRGB_NONLINEAR_KHR)
			swapchainImageFormat = aFormat;
   }

   // Retrieve supported presentation modes for the surface.
   uint32_t surfPresentModesCount;
   vkGetPhysicalDeviceSurfacePresentModesKHR(physicalDevice, surface, &surfPresentModesCount, NULL);
   vector<VkPresentModeKHR> surfPresentModes(surfPresentModesCount);
   vkGetPhysicalDeviceSurfacePresentModesKHR(physicalDevice, surface, &surfPresentModesCount, surfPresentModes.data());

   // Choose a presentation mode for the swapchain images:
   // Presentation mode VK_PRESENT_MODE_MAILBOX_KHR can be used to implement triple buffering,
   // so we'll look for this mode; in case we can't find it we'll go with 
   // VK_PRESENT_MODE_FIFO_KHR which is guaranteed to be supported.
   VkPresentModeKHR swapchainPresentMode;
   swapchainPresentMode = VK_PRESENT_MODE_FIFO_KHR;
   for (const auto& aPresentMode : surfPresentModes)
	  if (aPresentMode == VK_PRESENT_MODE_MAILBOX_KHR)
		 swapchainPresentMode = aPresentMode;

   // Get surface capabilities.
   VkSurfaceCapabilitiesKHR surfCapabilities;
   vkGetPhysicalDeviceSurfaceCapabilitiesKHR(physicalDevice, surface, &surfCapabilities);

   // Choose the extent (i.e., size) of the swapchain images:
   // If the current width and height of the surface are not the special value -1 (they are either both -1
   // or neither is) then we set the swapchain image size to the surface's current size.
   if (surfCapabilities.currentExtent.width != -1) swapchainImageExtent = surfCapabilities.currentExtent;

   // On the other hand, the width and height of the surface being defined -1 means that the size of the surface
   // is determined by its swapchain size rather than being fixed. In which case, we choose the
   // best values for the swapchain size keeping in mind that the globals windowWidth and windowHeight are our
   // preferences, but we are constrained to clamp width and height values between their respective minimum
   // and maximum for the surface's swapchain.
   else
   {
	  if (windowWidth < surfCapabilities.minImageExtent.width) swapchainImageExtent.width = surfCapabilities.minImageExtent.width;
	  else if (windowWidth > surfCapabilities.maxImageExtent.width) swapchainImageExtent.width = surfCapabilities.maxImageExtent.width;
	  else swapchainImageExtent.width = windowWidth;

	  if (windowHeight < surfCapabilities.minImageExtent.height) swapchainImageExtent.height = surfCapabilities.minImageExtent.height;
	  else if (windowHeight> surfCapabilities.maxImageExtent.height) swapchainImageExtent.height = surfCapabilities.maxImageExtent.height;
	  else swapchainImageExtent.height = windowHeight;
   }

   // Choose the minimum number of images in the swapchain:
   // We want to implement triple buffering so 3 images is our choice for minimum, but we our constrained to
   // lie within the minimum and maximum number of images supported for the surface's swapchain.
   // Note that the logic below takes into account that a maxImageCount values of 0 means there is no fixed 
   // maximum number of images.
   int swapchainMinImageCount = 3;
   if (swapchainMinImageCount < surfCapabilities.minImageCount) swapchainMinImageCount = surfCapabilities.minImageCount;
   else if (surfCapabilities.maxImageCount != 0 && swapchainMinImageCount > surfCapabilities.maxImageCount)
	  swapchainMinImageCount = surfCapabilities.maxImageCount;

   // Plug our choices above into swapchain info.
   VkSwapchainCreateInfoKHR createInfo = {};
   createInfo.sType = VK_STRUCTURE_TYPE_SWAPCHAIN_CREATE_INFO_KHR;
   createInfo.surface = surface;
   createInfo.minImageCount = swapchainMinImageCount;
   createInfo.imageFormat = swapchainImageFormat.format;
   createInfo.imageColorSpace = swapchainImageFormat.colorSpace;
   createInfo.imageExtent = swapchainImageExtent;
   createInfo.imageArrayLayers = 1;
   createInfo.imageUsage = VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT;
   createInfo.imageSharingMode = VK_SHARING_MODE_EXCLUSIVE;
   createInfo.preTransform = surfCapabilities.currentTransform;
   createInfo.compositeAlpha = VK_COMPOSITE_ALPHA_OPAQUE_BIT_KHR;
   createInfo.presentMode = swapchainPresentMode;
   createInfo.clipped = VK_TRUE;
   createInfo.oldSwapchain = VK_NULL_HANDLE;

   // Creagte swapchain
   vkCreateSwapchainKHR(logicalDevice, &createInfo, NULL, &swapchain);

   // Next, we obtain handles to images in the swapchain.
   // Note that the actual number of images created in the swapchain may exceed the 
   // minimum value swapchainMinImageCount we asked; therefore, we have to query
   // the exact number of images in the swapchain.
   uint32_t swapchainImageCount;
   vkGetSwapchainImagesKHR(logicalDevice, swapchain, &swapchainImageCount, NULL);
   swapchainImages.resize(swapchainImageCount);
   vkGetSwapchainImagesKHR(logicalDevice, swapchain, &swapchainImageCount, swapchainImages.data());

   // Create image views, one per image.
   swapchainImageViews.resize(swapchainImages.size(), VkImageView{});
   for (int i = 0; i < swapchainImages.size(); i++)
   {
	  // Image view info.
	  VkImageViewCreateInfo createInfo = {};
	  createInfo.sType = VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO;
	  createInfo.image = swapchainImages[i];
	  createInfo.viewType = VK_IMAGE_VIEW_TYPE_2D;
	  createInfo.format = swapchainImageFormat.format;
	  createInfo.components.r = VK_COMPONENT_SWIZZLE_R;
	  createInfo.components.g = VK_COMPONENT_SWIZZLE_G;
	  createInfo.components.b = VK_COMPONENT_SWIZZLE_B;
	  createInfo.components.a = VK_COMPONENT_SWIZZLE_A;
	  createInfo.subresourceRange.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
	  createInfo.subresourceRange.baseMipLevel = 0;
	  createInfo.subresourceRange.levelCount = 1;
	  createInfo.subresourceRange.baseArrayLayer = 0;
	  createInfo.subresourceRange.layerCount = 1;

	  // Create image views.
	  vkCreateImageView(logicalDevice, &createInfo, NULL, &swapchainImageViews[i]);
   }
}

// Function to read external SPIR-V file, modifed from readShader() of Chapter 15 to additionally
// output the filesize to an external variable and to not add a closing NUL to the output file. 
char* readSpirVfile(string fileName, int* fileSize)
{
   // Initialize input stream.
   ifstream inFile(fileName.c_str(), ios::binary);

   // Determine shader file length and reserve space to read it in.
   inFile.seekg(0, ios::end);
   int fileLength = inFile.tellg();
   *fileSize = fileLength;
   char *fileContent = (char*)malloc(fileLength * sizeof(char));

   // Read in shader file and close input stream.
   inFile.seekg(0, ios::beg);
   inFile.read(fileContent, fileLength);
   inFile.close();

   return fileContent;
}

// Create a handle to the shader module object made from a given SPIR-V file.
VkShaderModule setShaderModule(char* spirVFile)
{
   VkShaderModule shaderModule;
   int fileSize;
   char* spirvFile = readSpirVfile(spirVFile, &fileSize);

   // Shader module info.
   VkShaderModuleCreateInfo createInfo = {};
   createInfo.sType = VK_STRUCTURE_TYPE_SHADER_MODULE_CREATE_INFO;
   createInfo.codeSize = fileSize;
   createInfo.pCode = (uint32_t*)spirvFile;

   // Create shader module.
   vkCreateShaderModule(logicalDevice, &createInfo, NULL, &shaderModule);

   return shaderModule;
}

// Create a graphics pipeline filling in the member variables and structures of a
// VkGraphicsPipelineCreateInfo object. We do this "bottom-up" with the member 
// structures set first and finally the VkGraphicsPipelineCreateInfo object.
void createGraphicsPipeline()
{
   // Create shader modules.
   VkShaderModule vertShaderModule = setShaderModule("vertexShader.spv");
   VkShaderModule fragShaderModule = setShaderModule("fragmentShader.spv");

   // Vertex shader stage create info.
   VkPipelineShaderStageCreateInfo vertShaderStageInfo = {};
   vertShaderStageInfo.sType = VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_CREATE_INFO;
   vertShaderStageInfo.stage = VK_SHADER_STAGE_VERTEX_BIT;
   vertShaderStageInfo.module = vertShaderModule;
   vertShaderStageInfo.pName = "main";

   // Fragment shader stage create info.
   VkPipelineShaderStageCreateInfo fragShaderStageInfo = {};
   fragShaderStageInfo.sType = VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_CREATE_INFO;
   fragShaderStageInfo.stage = VK_SHADER_STAGE_FRAGMENT_BIT;
   fragShaderStageInfo.module = fragShaderModule;
   fragShaderStageInfo.pName = "main";

   // Vector of shader stage create infos.
   vector<VkPipelineShaderStageCreateInfo> shaderStageInfo = { vertShaderStageInfo, fragShaderStageInfo };

   // Vertex input buffer info.
   VkVertexInputBindingDescription vertexBufferBinding = {};
   vertexBufferBinding.binding = 0;
   vertexBufferBinding.stride = sizeof(Vertex);
   vertexBufferBinding.inputRate = VK_VERTEX_INPUT_RATE_VERTEX;

   // Vector of vertex input buffer infos (only one).
   vector<VkVertexInputBindingDescription> bindingDescriptions = { vertexBufferBinding };

   // Coords attribute info.
   VkVertexInputAttributeDescription coordsAttrib = {};
   coordsAttrib.binding = 0;
   coordsAttrib.location = 0;
   coordsAttrib.format = VK_FORMAT_R32G32B32A32_SFLOAT;
   coordsAttrib.offset = offsetof(Vertex, coords);

   // Color attribute info.
   VkVertexInputAttributeDescription colorsAttrib = {};
   colorsAttrib.binding = 0;
   colorsAttrib.location = 1;
   colorsAttrib.format = VK_FORMAT_R32G32B32A32_SFLOAT;
   colorsAttrib.offset = offsetof(Vertex, colors);

   // Vector of input attribute infos.
   vector<VkVertexInputAttributeDescription> attributeDescriptions = { coordsAttrib, colorsAttrib };

   // Pipeline vertex input info.
   VkPipelineVertexInputStateCreateInfo vertexInputInfo = {};
   vertexInputInfo.sType = VK_STRUCTURE_TYPE_PIPELINE_VERTEX_INPUT_STATE_CREATE_INFO;
   vertexInputInfo.vertexBindingDescriptionCount = bindingDescriptions.size();
   vertexInputInfo.pVertexBindingDescriptions = bindingDescriptions.data();
   vertexInputInfo.vertexAttributeDescriptionCount = attributeDescriptions.size();
   vertexInputInfo.pVertexAttributeDescriptions = attributeDescriptions.data();

   // Input assembly info.
   VkPipelineInputAssemblyStateCreateInfo inputAssemblyInfo = {};
   inputAssemblyInfo.sType = VK_STRUCTURE_TYPE_PIPELINE_INPUT_ASSEMBLY_STATE_CREATE_INFO;
   inputAssemblyInfo.topology = VK_PRIMITIVE_TOPOLOGY_TRIANGLE_STRIP;
   inputAssemblyInfo.primitiveRestartEnable = VK_FALSE;

   // Viewport info.
   VkViewport viewportInfo = {};
   viewportInfo.x = 0.0;
   viewportInfo.y = 0.0;
   viewportInfo.width = swapchainImageExtent.width;
   viewportInfo.height = swapchainImageExtent.height;
   viewportInfo.minDepth = 0.0;
   viewportInfo.maxDepth = 1.0;

   // Scissor info.
   VkRect2D scissorInfo = {};
   scissorInfo.offset = { 0, 0 };
   scissorInfo.extent = swapchainImageExtent;

   // ViewportState info = viewports + scissors info (since we'll always have one viewport
   // and one scissor we don't create a vector for either).
   VkPipelineViewportStateCreateInfo viewportStateInfo = {};
   viewportStateInfo.sType = VK_STRUCTURE_TYPE_PIPELINE_VIEWPORT_STATE_CREATE_INFO;
   viewportStateInfo.viewportCount = 1;
   viewportStateInfo.pViewports = &viewportInfo;
   viewportStateInfo.scissorCount = 1;
   viewportStateInfo.pScissors = &scissorInfo;

   // Rasterization info (options re depth are left at default as we don't depth test).
   VkPipelineRasterizationStateCreateInfo rasterizationInfo = {};
   rasterizationInfo.sType = VK_STRUCTURE_TYPE_PIPELINE_RASTERIZATION_STATE_CREATE_INFO;
   rasterizationInfo.rasterizerDiscardEnable = VK_FALSE;
   rasterizationInfo.polygonMode = VK_POLYGON_MODE_FILL;
   rasterizationInfo.cullMode = VK_CULL_MODE_NONE;
   rasterizationInfo.frontFace = VK_FRONT_FACE_COUNTER_CLOCKWISE;
   rasterizationInfo.lineWidth = 1.0; // Note that to draw lines wider than 1.0, wideLines 
									  // physical device feature must be enabled. Ditto for
									  // some other pipeline features. See Vulkan specs.

   // Multisampling info (we don't multisample and leave further options at default).
   VkPipelineMultisampleStateCreateInfo multisampleInfo = {};
   multisampleInfo.sType = VK_STRUCTURE_TYPE_PIPELINE_MULTISAMPLE_STATE_CREATE_INFO;
   multisampleInfo.rasterizationSamples = VK_SAMPLE_COUNT_1_BIT;

   // Depth and stencil testing info (we do neither and leave further options at default).
   VkPipelineDepthStencilStateCreateInfo depthStencilInfo = {};
   depthStencilInfo.sType = VK_STRUCTURE_TYPE_PIPELINE_DEPTH_STENCIL_STATE_CREATE_INFO;
   depthStencilInfo.depthTestEnable = VK_FALSE;
   depthStencilInfo.stencilTestEnable = VK_FALSE;

   // Color attachment blend info 
   // (we have only one color attachment which we draw and which is the output).
   VkPipelineColorBlendAttachmentState colorBlendAttachment = {};
   colorBlendAttachment.blendEnable = VK_FALSE;
   colorBlendAttachment.colorWriteMask = VK_COLOR_COMPONENT_R_BIT | VK_COLOR_COMPONENT_G_BIT | VK_COLOR_COMPONENT_B_BIT | VK_COLOR_COMPONENT_A_BIT;

   // Vector of color attachments.
   vector<VkPipelineColorBlendAttachmentState> colorBlendAttachments = { colorBlendAttachment };

   // Color blending info.
   VkPipelineColorBlendStateCreateInfo colorBlendInfo = {};
   colorBlendInfo.sType = VK_STRUCTURE_TYPE_PIPELINE_COLOR_BLEND_STATE_CREATE_INFO;
   colorBlendInfo.logicOpEnable = VK_FALSE;
   colorBlendInfo.attachmentCount = colorBlendAttachments.size();
   colorBlendInfo.pAttachments = colorBlendAttachments.data();

   // Pipeline layout info (we have no descriptor sets or push constants).
   VkPipelineLayoutCreateInfo pipelineLayoutInfo = {};
   pipelineLayoutInfo.sType = VK_STRUCTURE_TYPE_PIPELINE_LAYOUT_CREATE_INFO;
   pipelineLayoutInfo.setLayoutCount = 1;
   pipelineLayoutInfo.pSetLayouts = &descriptorSetLayout;

   // Create layout.
   vkCreatePipelineLayout(logicalDevice, &pipelineLayoutInfo, NULL, &pipelineLayout);

   // Pipeline info.
   VkGraphicsPipelineCreateInfo pipelineInfo = {};
   pipelineInfo.sType = VK_STRUCTURE_TYPE_GRAPHICS_PIPELINE_CREATE_INFO;
   pipelineInfo.flags = VK_PIPELINE_CREATE_DISABLE_OPTIMIZATION_BIT;
   pipelineInfo.stageCount = shaderStageInfo.size();
   pipelineInfo.pStages = shaderStageInfo.data();
   pipelineInfo.pVertexInputState = &vertexInputInfo;
   pipelineInfo.pInputAssemblyState = &inputAssemblyInfo;
   pipelineInfo.pViewportState = &viewportStateInfo;
   pipelineInfo.pRasterizationState = &rasterizationInfo;
   pipelineInfo.pMultisampleState = &multisampleInfo;
   pipelineInfo.pDepthStencilState = &depthStencilInfo;
   pipelineInfo.pColorBlendState = &colorBlendInfo;
   pipelineInfo.pDynamicState = NULL;
   pipelineInfo.layout = pipelineLayout;
   pipelineInfo.renderPass = renderpass; // This pipeline operates for this renderpass ...
   pipelineInfo.subpass = 0; // ... at this  particular subpass.
   pipelineInfo.basePipelineHandle = VK_NULL_HANDLE;

   // Create pipeline.
   vkCreateGraphicsPipelines(logicalDevice, VK_NULL_HANDLE, 1, &pipelineInfo, NULL, &graphicsPipeline);
}

// Create a renderpass for our graphics pipeline.
void createRenderpass()
{
   // We  have one attachment, a color attachment.
   VkAttachmentDescription colorAttachment = {};
   colorAttachment.format = swapchainImageFormat.format;
   colorAttachment.samples = VK_SAMPLE_COUNT_1_BIT;
   colorAttachment.loadOp = VK_ATTACHMENT_LOAD_OP_CLEAR;
   colorAttachment.storeOp = VK_ATTACHMENT_STORE_OP_STORE;
   colorAttachment.stencilLoadOp = VK_ATTACHMENT_LOAD_OP_DONT_CARE;
   colorAttachment.stencilStoreOp = VK_ATTACHMENT_STORE_OP_DONT_CARE;
   colorAttachment.initialLayout = VK_IMAGE_LAYOUT_UNDEFINED;
   colorAttachment.finalLayout = VK_IMAGE_LAYOUT_PRESENT_SRC_KHR;

   // Vector of attachments.
   vector<VkAttachmentDescription> attachments = {colorAttachment};

   // The one subpass refers and writes to the one color attachment.
   VkAttachmentReference attachmentRef = {};
   attachmentRef.attachment = 0;
   attachmentRef.layout = VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL;

   // Vector of attachment references.
   vector<VkAttachmentReference> attachmentRefs = {attachmentRef};

   // Subpass info.
   VkSubpassDescription subpass = {};
   subpass.pipelineBindPoint = VK_PIPELINE_BIND_POINT_GRAPHICS;
   subpass.colorAttachmentCount = attachmentRefs.size();
   subpass.pColorAttachments = attachmentRefs.data();

   // Vector of subpass infos.
   vector<VkSubpassDescription> subpasses = {subpass};

   // Renderpass info.
   VkRenderPassCreateInfo renderPassInfo = {};
   renderPassInfo.sType = VK_STRUCTURE_TYPE_RENDER_PASS_CREATE_INFO;
   renderPassInfo.attachmentCount = attachments.size();
   renderPassInfo.pAttachments = attachments.data();
   renderPassInfo.subpassCount = subpasses.size();
   renderPassInfo.pSubpasses = subpasses.data();

   // Create renderpass.
   vkCreateRenderPass(logicalDevice, &renderPassInfo, NULL, &renderpass);
}

// Create framebuffers.
void createFramebuffers()
{
   // One frame buffer per swapchain image.
   framebuffers.resize(swapchainImages.size(), VkFramebuffer{});

   for (int i = 0; i < swapchainImages.size(); i++)
   {
	  // Vector of framebuffer attachments (only one here).
	  vector<VkImageView> attachments = {swapchainImageViews[i]};

	  // Framebuffer info.
	  VkFramebufferCreateInfo framebufferInfo = {};
	  framebufferInfo.sType = VK_STRUCTURE_TYPE_FRAMEBUFFER_CREATE_INFO;
	  framebufferInfo.renderPass = renderpass;
	  framebufferInfo.attachmentCount = attachments.size();
	  framebufferInfo.pAttachments = attachments.data();
	  framebufferInfo.width = swapchainImageExtent.width;
	  framebufferInfo.height = swapchainImageExtent.height;
	  framebufferInfo.layers = 1;

	  // Create framebuffer.
	  vkCreateFramebuffer(logicalDevice, &framebufferInfo, NULL, &framebuffers[i]);
   }
}

// Create two command pools.
void createCommandPools()
{
   // Command pool create info.
   VkCommandPoolCreateInfo cmdPoolInfo = {};
   cmdPoolInfo.sType = VK_STRUCTURE_TYPE_COMMAND_POOL_CREATE_INFO;
   cmdPoolInfo.queueFamilyIndex = queueFamily;

   // Create command pool for the draw command buffer.
   vkCreateCommandPool(logicalDevice, &cmdPoolInfo, NULL, &commandPoolDraw);

   // Create command pool for the copy command buffer, this buffer being transient.
   cmdPoolInfo.flags = VK_COMMAND_POOL_CREATE_TRANSIENT_BIT;
   vkCreateCommandPool(logicalDevice, &cmdPoolInfo, NULL, &commandPoolCopy);
}

// Allocate draw command buffers and record commands.
void createDrawCommandBuffers()
{
   // Set one draw command buffer per framebuffer.
   commandBuffersDraw.resize(framebuffers.size());

   // Command buffer info.  
   VkCommandBufferAllocateInfo cmdBufferAllocInfo = {};
   cmdBufferAllocInfo.sType = VK_STRUCTURE_TYPE_COMMAND_BUFFER_ALLOCATE_INFO;
   cmdBufferAllocInfo.commandPool = commandPoolDraw;
   cmdBufferAllocInfo.level = VK_COMMAND_BUFFER_LEVEL_PRIMARY;
   cmdBufferAllocInfo.commandBufferCount = commandBuffersDraw.size();

   // Allocate command buffers for draw commands.
   vkAllocateCommandBuffers(logicalDevice, &cmdBufferAllocInfo, commandBuffersDraw.data());

   // Record drawing commands in each command buffer.
   for (int i = 0; i < commandBuffersDraw.size(); i++)
   {
	  // Command buffer begin info.
	  VkCommandBufferBeginInfo cmdBufferBeginInfo = {};
	  cmdBufferBeginInfo.sType = VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO;
	  cmdBufferBeginInfo.flags = VK_COMMAND_BUFFER_USAGE_SIMULTANEOUS_USE_BIT;

	  // Clear color.
	  VkClearValue clearColor = {1.0, 1.0, 1.0, 1.0};

	  // Vector of clear colors (only one as we have one attachment).
	  vector<VkClearValue> clearColors = { clearColor };

	  // Renderpass begin info.
	  VkRenderPassBeginInfo renderPassBeginInfo = {};
	  renderPassBeginInfo.sType = VK_STRUCTURE_TYPE_RENDER_PASS_BEGIN_INFO;
	  renderPassBeginInfo.renderPass = renderpass;
	  renderPassBeginInfo.framebuffer = framebuffers[i];
	  renderPassBeginInfo.renderArea.offset = { 0, 0 };
	  renderPassBeginInfo.renderArea.extent = swapchainImageExtent;
	  renderPassBeginInfo.clearValueCount = clearColors.size();
	  renderPassBeginInfo.pClearValues = clearColors.data();

	  // Set vertex buffers and offsets.
	  VkBuffer vertexBuffers[] = { vertexBuffer };
	  VkDeviceSize offsets[] = { 0 };

	  // Record commands.
	  vkBeginCommandBuffer(commandBuffersDraw[i], &cmdBufferBeginInfo);
	  vkCmdBeginRenderPass(commandBuffersDraw[i], &renderPassBeginInfo, VK_SUBPASS_CONTENTS_INLINE);
	  vkCmdBindPipeline(commandBuffersDraw[i], VK_PIPELINE_BIND_POINT_GRAPHICS, graphicsPipeline);
	  vkCmdBindVertexBuffers(commandBuffersDraw[i], 0, 1, vertexBuffers, offsets);
	  vkCmdBindDescriptorSets(commandBuffersDraw[i], VK_PIPELINE_BIND_POINT_GRAPHICS, pipelineLayout, 0, 1, &descriptorSet, 0, NULL);
	  vkCmdDraw(commandBuffersDraw[i], sizeof(squareVertices)/sizeof(squareVertices[0]), 1, 0, 0);
	  vkCmdEndRenderPass(commandBuffersDraw[i]);
	  vkEndCommandBuffer(commandBuffersDraw[i]);
   }
}

// Create fence and semaphore.
void createSynchronizingPrimitives()
{
   // Fence info.
   VkFenceCreateInfo fenceInfo = {};
   fenceInfo.sType = VK_STRUCTURE_TYPE_FENCE_CREATE_INFO;

   // Create fence.
   vkCreateFence(logicalDevice, &fenceInfo, NULL, &imageAcquiredFence);

   // Semaphore info.
   VkSemaphoreCreateInfo semaphoreInfo = {};
   semaphoreInfo.sType = VK_STRUCTURE_TYPE_SEMAPHORE_CREATE_INFO;

   // Create semaphore.
   vkCreateSemaphore(logicalDevice, &semaphoreInfo, NULL, &renderCompleteSemaphore);
}

// Draw a frame.
void drawFrame()
{
   uint32_t imageIndex;

   // Acquire image and signal fence.
   vkAcquireNextImageKHR(logicalDevice, swapchain, UINT64_MAX, VK_NULL_HANDLE, imageAcquiredFence, &imageIndex);

   // Wait for fence to be signalled 
   vkWaitForFences(logicalDevice, 1, &imageAcquiredFence, VK_TRUE, UINT64_MAX);

   // Reset the fence after it is signalled, i.e., prepare for the next cycle.
   vkResetFences(logicalDevice, 1, &imageAcquiredFence);

   // Queue submit info.
   VkSubmitInfo submitInfo = {};
   submitInfo.sType = VK_STRUCTURE_TYPE_SUBMIT_INFO;
   submitInfo.commandBufferCount = 1;
   submitInfo.pCommandBuffers = &commandBuffersDraw[imageIndex];
   submitInfo.signalSemaphoreCount = 1;
   submitInfo.pSignalSemaphores = &renderCompleteSemaphore;

   // Submit the queue commands.
   if (vkQueueSubmit(queue, 1, &submitInfo, VK_NULL_HANDLE) != VK_SUCCESS) {
	  throw runtime_error("failed to submit draw command buffer!");
   }

   // Presentation (to surface) info.
   VkPresentInfoKHR presentInfo = {};
   presentInfo.sType = VK_STRUCTURE_TYPE_PRESENT_INFO_KHR;
   presentInfo.waitSemaphoreCount = 1;
   presentInfo.pWaitSemaphores = &renderCompleteSemaphore;
   presentInfo.swapchainCount = 1;
   presentInfo.pSwapchains = &swapchain;
   presentInfo.pImageIndices = &imageIndex;

   // Queue the image for presentation.
   vkQueuePresentKHR(queue, &presentInfo);
}

// Choose appropriate memory type for a buffer given additional properties asked of the memory.
int chooseMemoryType(int supportingTypes, // Memory types fitting the buffer's requirements.
   VkMemoryPropertyFlags askedProperties // Asked memory properties.
)
{
   VkPhysicalDeviceMemoryProperties memProperties;

   // Get available memory types.
   vkGetPhysicalDeviceMemoryProperties(physicalDevice, &memProperties);

   // Run through available memory types looking for a fit.
   for (int i = 0; i < memProperties.memoryTypeCount; i++)
   {
	  if ((supportingTypes & (1 << i)) // True if memory type i supports the buffer.
		 && ((memProperties.memoryTypes[i].propertyFlags & askedProperties) == askedProperties) // True if memory has desired properties.
		 )
		 return i;
   }
   cout << "Failed to find memory!" << endl;
   exit(EXIT_FAILURE);
}

// Generic routine to create a buffer with given size for given usage 
// and bind to it memory with additional asked properties.
void createBufferAndBindMemory(VkDeviceSize bufferSize, // Given buffer size. 
                               VkBufferUsageFlags bufferUsage, // Given buffer use.
                               VkMemoryPropertyFlags askedProperties, // Specified memory properties.
                               VkBuffer& buffer, // Pointer to created buffer.
                               VkDeviceMemory& bufferMemory) // Pointer to bound memory.
{
   // Buffer info.
   VkBufferCreateInfo bufferInfo = {};
   bufferInfo.sType = VK_STRUCTURE_TYPE_BUFFER_CREATE_INFO;
   bufferInfo.size = bufferSize;
   bufferInfo.usage = bufferUsage;
   bufferInfo.sharingMode = VK_SHARING_MODE_EXCLUSIVE;

   // Create buffer.
   vkCreateBuffer(logicalDevice, &bufferInfo, NULL, &buffer);

   // Get buffer memory requirements.
   VkMemoryRequirements memRequirements;
   vkGetBufferMemoryRequirements(logicalDevice, buffer, &memRequirements);

   // Memory info.
   VkMemoryAllocateInfo memAllocInfo = {};
   memAllocInfo.sType = VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO;
   memAllocInfo.allocationSize = memRequirements.size;
   memAllocInfo.memoryTypeIndex = chooseMemoryType(memRequirements.memoryTypeBits, askedProperties);

   // Allocate memory.
   vkAllocateMemory(logicalDevice, &memAllocInfo, NULL, &bufferMemory);

   // Bind memory to buffer.
   vkBindBufferMemory(logicalDevice, buffer, bufferMemory, 0);
}

// Create a vertex buffer and its corresponding staging buffer and their respective bound memories.
void createVertexAndStagingBuffers()
{
   // Create buffer for use as vertex buffer and as memory transfer destination; 
   // bind to it memory which is GPU local (for fast reads during run time).
   createBufferAndBindMemory(sizeof(squareVertices),
	  VK_BUFFER_USAGE_VERTEX_BUFFER_BIT | VK_BUFFER_USAGE_TRANSFER_DST_BIT,
	  VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT,
	  vertexBuffer, vertexBufferMemory);

   // Create buffer for use as memory transfer source; bind to it memory which
   // can be mapped by the host system (so that vertex data can be copied to it
   // from the host). This buffer will serve as a staging (intermediate) buffer
   // to which memory will be copied from the host and then transferred to the 
   // vertex buffer.
   createBufferAndBindMemory(sizeof(squareVertices),
	  VK_BUFFER_USAGE_TRANSFER_SRC_BIT,
	  VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT | VK_MEMORY_PROPERTY_HOST_COHERENT_BIT,
	  vertexStagingBuffer, vertexStagingBufferMemory);
}

// Copy data from host to staging buffer and then to vertex buffer.
void moveDataToVertexBuffer()
{
   void* data;

   // Map vertex staging buffer memory and copy in data from host.
   vkMapMemory(logicalDevice, vertexStagingBufferMemory, 0, sizeof(squareVertices), 0, &data);
   memcpy(data, squareVertices, sizeof(squareVertices));
   vkUnmapMemory(logicalDevice, vertexStagingBufferMemory);

   // Command buffer info.
   VkCommandBufferAllocateInfo cmdBufferAllocInfo = {};
   cmdBufferAllocInfo.sType = VK_STRUCTURE_TYPE_COMMAND_BUFFER_ALLOCATE_INFO;
   cmdBufferAllocInfo.level = VK_COMMAND_BUFFER_LEVEL_PRIMARY;
   cmdBufferAllocInfo.commandPool = commandPoolCopy;
   cmdBufferAllocInfo.commandBufferCount = 1;

   // Allocate command buffer for copy instruction.
   vkAllocateCommandBuffers(logicalDevice, &cmdBufferAllocInfo, &commandBufferCopy);

   // Command buffer begin info.
   VkCommandBufferBeginInfo cmdBufferBeginInfo = {};
   cmdBufferBeginInfo.sType = VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO;
   cmdBufferBeginInfo.flags = VK_COMMAND_BUFFER_USAGE_ONE_TIME_SUBMIT_BIT;

   // Begin recording command buffer.
   vkBeginCommandBuffer(commandBufferCopy, &cmdBufferBeginInfo);

   // Parameters for copy command.
   VkBufferCopy copyParams = {};

   // Set size of vertex data and copy it from staging to vertex buffer.
   copyParams.size = sizeof(squareVertices);
   vkCmdCopyBuffer(commandBufferCopy, vertexStagingBuffer, vertexBuffer, 1, &copyParams);

   // End recording commands.
   vkEndCommandBuffer(commandBufferCopy);

   // Queue submit info.
   VkSubmitInfo submitInfo = {};
   submitInfo.sType = VK_STRUCTURE_TYPE_SUBMIT_INFO;
   submitInfo.commandBufferCount = 1;
   submitInfo.pCommandBuffers = &commandBufferCopy;

   // Submit the queue commands.
   vkQueueSubmit(queue, 1, &submitInfo, VK_NULL_HANDLE);

   // When the queue is idle release staging buffers and memory.
   vkQueueWaitIdle(queue);
   vkFreeCommandBuffers(logicalDevice, commandPoolCopy, 1, &commandBufferCopy);
   vkDestroyBuffer(logicalDevice, vertexStagingBuffer, NULL);
   vkFreeMemory(logicalDevice, vertexStagingBufferMemory, NULL);
}

// Create a descriptor set layout.
void createDescriptorSetLayout()
{
   // Set layout has one binding for a uniform buffer.
   VkDescriptorSetLayoutBinding uniformBufferLayoutBinding = {};
   uniformBufferLayoutBinding.binding = 0;
   uniformBufferLayoutBinding.descriptorType = VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER;
   uniformBufferLayoutBinding.descriptorCount = 1;
   uniformBufferLayoutBinding.stageFlags = VK_SHADER_STAGE_VERTEX_BIT;

   // Descriptor set layout info.
   VkDescriptorSetLayoutCreateInfo descSetLayoutInfo = {};
   descSetLayoutInfo.sType = VK_STRUCTURE_TYPE_DESCRIPTOR_SET_LAYOUT_CREATE_INFO;
   descSetLayoutInfo.bindingCount = 1;
   descSetLayoutInfo.pBindings = &uniformBufferLayoutBinding;

   // Create descriptor set layout.
   vkCreateDescriptorSetLayout(logicalDevice, &descSetLayoutInfo, NULL, &descriptorSetLayout);
}

// Create uniform buffer and its bound memory which is mapped by the host system
// (for fast writing of uniforms which may change per frame).
void createUniformBuffer()
{
   VkDeviceSize bufferSize = sizeof(matrixUniforms);
   createBufferAndBindMemory(bufferSize, VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT,
	  VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT | VK_MEMORY_PROPERTY_HOST_COHERENT_BIT,
	  uniformBuffer, uniformBufferMemory);
}

// Create descriptor pool. 
void createDescriptorPool() 
{
   // One descriptor pool size structure corresponding to one binding with one uniform buffer.
   VkDescriptorPoolSize descPoolSize = {};
   descPoolSize.type = VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER;
   descPoolSize.descriptorCount = 1;

   // Descriptor pool info. 
   VkDescriptorPoolCreateInfo descPoolInfo = {};
   descPoolInfo.sType = VK_STRUCTURE_TYPE_DESCRIPTOR_POOL_CREATE_INFO;
   descPoolInfo.maxSets = 1; // One descriptor set.
   descPoolInfo.poolSizeCount = 1; // One pool size structure.
   descPoolInfo.pPoolSizes = &descPoolSize;

   // Create descriptor pool.
   vkCreateDescriptorPool(logicalDevice, &descPoolInfo, NULL, &descriptorPool);
}

// Allocate descriptor set, attach memory to the one uniform descriptor and update the descriptor.
void createAndUpdateDescriptorSet() 
{
   // Allocate info (one descriptor set).
   VkDescriptorSetAllocateInfo descSetAllocInfo = {};
   descSetAllocInfo.sType = VK_STRUCTURE_TYPE_DESCRIPTOR_SET_ALLOCATE_INFO;
   descSetAllocInfo.descriptorPool = descriptorPool;
   descSetAllocInfo.descriptorSetCount = 1;
   descSetAllocInfo.pSetLayouts = &descriptorSetLayout;

   // Allocate descriptor set from descriptor pool.
   vkAllocateDescriptorSets(logicalDevice, &descSetAllocInfo, &descriptorSet);

   // Uniform descriptor buffer info.
   VkDescriptorBufferInfo bufferInfo = {};
   bufferInfo.buffer = uniformBuffer;
   bufferInfo.offset = 0;
   bufferInfo.range = sizeof(matrixUniforms);

   // Info to map data from uniform buffer to uniform descriptor.
   VkWriteDescriptorSet descriptorWrite = {};
   descriptorWrite.sType = VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET;
   descriptorWrite.dstSet = descriptorSet;
   descriptorWrite.dstBinding = 0;
   descriptorWrite.dstArrayElement = 0;
   descriptorWrite.descriptorCount = 1;
   descriptorWrite.descriptorType = VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER;
   descriptorWrite.pBufferInfo = &bufferInfo;

   // Update data in uniform descriptor per descriptorWrite info above.
   vkUpdateDescriptorSets(logicalDevice, 1, &descriptorWrite, 0, NULL);
}

// Set uniform matrices data in host and copy to uniform buffer.
void moveDataToUniformBuffer() 
{
   matrixUniforms.modelviewMat = mat4(1.0); // Identity modelview transformation.
   matrixUniforms.projMat = ortho(0.0, 100.0, 0.0, 100.0, -1.0, 1.0); // OpenGL: glOrtho(0.0, 100.0, 0.0, 100.0, -1.0, 1.0).

   // Map uniform buffer memory and copy in data from host.
   void* data;
   vkMapMemory(logicalDevice, uniformBufferMemory, 0, sizeof(matrixUniforms), 0, &data);
   memcpy(data, &matrixUniforms, sizeof(matrixUniforms));
   vkUnmapMemory(logicalDevice, uniformBufferMemory);
}

// Clean up resources related to the current swapchain (prior creating a new swapchain).
void cleanupSwapChain() 
{
   for (int i = 0; i < framebuffers.size(); i++) vkDestroyFramebuffer(logicalDevice, framebuffers[i], NULL);
   vkFreeCommandBuffers(logicalDevice, commandPoolDraw, commandBuffersDraw.size(), commandBuffersDraw.data());
   vkDestroyPipeline(logicalDevice, graphicsPipeline, NULL);
   vkDestroyPipelineLayout(logicalDevice, pipelineLayout, NULL);
   vkDestroyRenderPass(logicalDevice, renderpass, NULL);
   for (int i = 0; i < swapchainImageViews.size(); i++) vkDestroyImageView(logicalDevice, swapchainImageViews[i], NULL);
   vkDestroySwapchainKHR(logicalDevice, swapchain, NULL);
}

// Destroy current swapchain and create new one.
void recreateSwapChain() 
{
   vkDeviceWaitIdle(logicalDevice);

   cleanupSwapChain();

   createSwapchainAndImageViews();
   createRenderpass();
   createGraphicsPipeline();
   createFramebuffers();
   createDrawCommandBuffers();
}

// Clean up resources prior to program termination.
void cleanup() 
{
   cleanupSwapChain();

   vkDestroyDescriptorPool(logicalDevice, descriptorPool, NULL);
   vkDestroyDescriptorSetLayout(logicalDevice, descriptorSetLayout, NULL);
   
   vkDestroyBuffer(logicalDevice, uniformBuffer, NULL);
   vkFreeMemory(logicalDevice, uniformBufferMemory, NULL);
   vkDestroyBuffer(logicalDevice, vertexBuffer, NULL);
   vkFreeMemory(logicalDevice, vertexBufferMemory, NULL);

   vkDestroySemaphore(logicalDevice, renderCompleteSemaphore, NULL);
   vkDestroyFence(logicalDevice, imageAcquiredFence, NULL);

   vkDestroyCommandPool(logicalDevice, commandPoolDraw, NULL);
   vkDestroyCommandPool(logicalDevice, commandPoolCopy, NULL);

   vkDestroyDevice(logicalDevice, NULL);
   vkDestroySurfaceKHR(instance, surface, NULL);
   vkDestroyInstance(instance, NULL);

   glfwTerminate();
}

// Initialization routine.
void setup(void)
{
   createInstance();
   createSurface();
   pickPhysicalDeviceAndQueueFamily();
   createLogicalDeviceAndPickQueue();
   createSwapchainAndImageViews();
   createRenderpass();
   createDescriptorSetLayout();
   createGraphicsPipeline();
   createFramebuffers();
   createCommandPools();
   createVertexAndStagingBuffers();
   moveDataToVertexBuffer();
   createUniformBuffer();
   moveDataToUniformBuffer();
   createDescriptorPool();
   createAndUpdateDescriptorSet();
   createDrawCommandBuffers();
   createSynchronizingPrimitives();
}

// Window reshape routine.
void resize(GLFWwindow* window, int w, int h)
{
   windowWidth = w; windowHeight = h; // Set current window dimensions.
   recreateSwapChain(); // Recreate swapchain.
}

// Keyboard input processing routine.
void keyInput(GLFWwindow* window, int key, int scancode, int action, int mods)
{
   if (key == GLFW_KEY_ESCAPE && action == GLFW_PRESS || GLFW_REPEAT)
	  glfwSetWindowShouldClose(window, GLFW_TRUE);
}

// Main routine.
int main(int argc, char **argv)
{
   glfwInit();

   glfwWindowHint(GLFW_CLIENT_API, GLFW_NO_API);

   window = glfwCreateWindow(windowWidth, windowHeight, "squareVulkanized.cpp", NULL, NULL);

   glfwSetWindowPos(window, 100, 100);
   glfwSetWindowSizeCallback(window, resize);
   glfwSetKeyCallback(window, keyInput);

   setup();

   while (!glfwWindowShouldClose(window))
   {
	  glfwPollEvents();
	  drawFrame();
   }

   vkDeviceWaitIdle(logicalDevice);
   cleanup();
}